# Python Version: 3.x
__package_name__ = 'online-judge-tools'
__author__ = 'Kimiyuki Onaka'
__email__ = 'kimiyuki95@gmail.com'
__license__ = 'MIT License'
__url__ = 'https://github.com/kmyk/online-judge-tools'
__version_info__ = (7, 3, 0, 'final', 0)
__version__ = '.'.join(map(str, __version_info__[:3]))
__description__ = 'Tools for online-judge services'
